@@ -52,6 +52,8 @@
 #include <QtWaylandClient/private/qwayland-qt-text-input-method-unstable-v1.h>
 #include <QtWaylandClient/private/qwayland-fractional-scale-v1.h>
 #include <QtWaylandClient/private/qwayland-viewporter.h>
+#include <QtWaylandClient/private/qwayland-cursor-shape-v1.h>
+#include <QtWaylandClient/private/qwayland-qt-toplevel-drag-v1.h>
 
 #include <QtCore/private/qcore_unix_p.h>
 
@@ -65,18 +67,6 @@
 
 #include <tuple> // for std::tie
 
-static void checkWaylandError(struct wl_display *display)
-{
-    int ecode = wl_display_get_error(display);
-    if ((ecode == EPIPE || ecode == ECONNRESET)) {
-        // special case this to provide a nicer error
-        qWarning("The Wayland connection broke. Did the Wayland compositor die?");
-    } else {
-        qWarning("The Wayland connection experienced a fatal error: %s", strerror(ecode));
-    }
-    _exit(1);
-}
-
 QT_BEGIN_NAMESPACE
 
 namespace QtWaylandClient {
@@ -114,9 +104,13 @@ public:
          * not only the one issued from event thread's waitForReading(), which means functions
          * called from dispatch_pending() can safely spin an event loop.
          */
+        if (m_quitting)
+            return;
+
         for (;;) {
             if (dispatchQueuePending() < 0) {
-                checkWaylandError(m_wldisplay);
+                Q_EMIT waylandError();
+                m_quitting = true;
                 return;
             }
 
@@ -154,6 +148,7 @@ public:
 
 Q_SIGNALS:
     void needReadAndDispatch();
+    void waylandError();
 
 protected:
     void run() override
@@ -284,7 +279,7 @@ struct ::wl_region *QWaylandDisplay::createRegion(const QRegion &qregion)
 
 ::wl_subsurface *QWaylandDisplay::createSubSurface(QWaylandWindow *window, QWaylandWindow *parent)
 {
-    if (!mSubCompositor) {
+    if (!mGlobals.subCompositor) {
         qCWarning(lcQpaWayland) << "Can't create subsurface, not supported by the compositor.";
         return nullptr;
     }
@@ -293,18 +288,18 @@ struct ::wl_region *QWaylandDisplay::createRegion(const QRegion &qregion)
     Q_ASSERT(parent->wlSurface());
     Q_ASSERT(window->wlSurface());
 
-    return mSubCompositor->get_subsurface(window->wlSurface(), parent->wlSurface());
+    return mGlobals.subCompositor->get_subsurface(window->wlSurface(), parent->wlSurface());
 }
 
 ::wp_viewport *QWaylandDisplay::createViewport(QWaylandWindow *window)
 {
-    if (!mViewporter) {
+    if (!mGlobals.viewporter) {
         qCWarning(lcQpaWayland) << "Can't create wp_viewport, not supported by the compositor.";
         return nullptr;
     }
 
     Q_ASSERT(window->wlSurface());
-    return mViewporter->get_viewport(window->wlSurface());
+    return mGlobals.viewporter->get_viewport(window->wlSurface());
 }
 
 QWaylandShellIntegration *QWaylandDisplay::shellIntegration() const
@@ -328,11 +323,17 @@ QWaylandDisplay::QWaylandDisplay(QWaylandIntegration *waylandIntegration)
     qRegisterMetaType<uint32_t>("uint32_t");
 
     mDisplay = wl_display_connect(nullptr);
-    if (!mDisplay) {
+    if (mDisplay) {
+        setupConnection();
+    } else {
         qErrnoWarning(errno, "Failed to create wl_display");
-        return;
     }
 
+    mWaylandTryReconnect = qEnvironmentVariableIsSet("QT_WAYLAND_RECONNECT");
+}
+
+void QWaylandDisplay::setupConnection()
+{
     struct ::wl_registry *registry = wl_display_get_registry(mDisplay);
     init(registry);
 
@@ -365,9 +366,6 @@ QWaylandDisplay::~QWaylandDisplay(void)
     }
     qDeleteAll(mWaitingScreens);
 
-#if QT_CONFIG(wayland_datadevice)
-    mDndSelectionHandler.reset();
-#endif
 #if QT_CONFIG(cursor)
     mCursorThemes.clear();
 #endif
@@ -375,6 +373,8 @@ QWaylandDisplay::~QWaylandDisplay(void)
     if (m_frameEventQueue)
         wl_event_queue_destroy(m_frameEventQueue);
 
+    mGlobals = {};
+
     if (mDisplay)
         wl_display_disconnect(mDisplay);
 }
@@ -410,7 +410,96 @@ void QWaylandDisplay::ensureScreen()
     Q_ASSERT(!QGuiApplication::screens().empty());
 }
 
-// Called in main thread, either from queued signal or directly.
+void QWaylandDisplay::reconnect()
+{
+    qCWarning(lcQpaWayland) << "Attempting wayland reconnect";
+    m_eventThread->stop();
+    m_frameEventQueueThread->stop();
+    m_eventThread->wait();
+    m_frameEventQueueThread->wait();
+
+    qDeleteAll(mWaitingScreens);
+    mWaitingScreens.clear();
+
+    while (!mScreens.isEmpty()) {
+        auto screen = mScreens.takeLast();
+        ensureScreen();
+        QWindowSystemInterface::handleScreenRemoved(screen);
+    }
+
+    // mCompositor
+    mShm.reset();
+    mCursorThemes.clear();
+    mCursor.reset();
+
+    mGlobals = GlobalHolder();
+
+    mWaylandIntegration->reset();
+
+    qDeleteAll(std::exchange(mInputDevices, {}));
+    mLastInputDevice = nullptr;
+
+    for (const RegistryGlobal &global : mRegistryGlobals) {
+        emit globalRemoved(global);
+    }
+    mRegistryGlobals.clear();
+
+    mLastInputSerial = 0;
+    mLastInputWindow.clear();
+    mLastKeyboardFocus.clear();
+    mActiveWindows.clear();
+
+    const auto windows = QGuiApplication::allWindows();
+    for (auto window : windows) {
+        if (auto waylandWindow = dynamic_cast<QWaylandWindow *>(window->handle()))
+            waylandWindow->closeChildPopups();
+    }
+    // Remove windows that do not need to be recreated and now closed popups
+    QList<QWaylandWindow *> recreateWindows;
+    for (auto window : std::as_const(windows)) {
+        auto waylandWindow = dynamic_cast<QWaylandWindow*>((window)->handle());
+        if (waylandWindow && waylandWindow->wlSurface()) {
+            waylandWindow->reset();
+            recreateWindows.push_back(waylandWindow);
+        }
+    }
+
+    if (mSyncCallback) {
+        wl_callback_destroy(mSyncCallback);
+        mSyncCallback = nullptr;
+    }
+
+    mDisplay = wl_display_connect(nullptr);
+    if (!mDisplay)
+        _exit(1);
+
+    setupConnection();
+    initialize();
+
+    if (m_frameEventQueue)
+        wl_event_queue_destroy(m_frameEventQueue);
+    initEventThread();
+
+    emit reconnected();
+
+    auto needsRecreate = [](QPlatformWindow *window) {
+        return window && !static_cast<QWaylandWindow *>(window)->wlSurface();
+    };
+    auto window = recreateWindows.begin();
+    while (!recreateWindows.isEmpty()) {
+        if (!needsRecreate((*window)->QPlatformWindow::parent()) && !needsRecreate((*window)->transientParent())) {
+            (*window)->reinit();
+            window = recreateWindows.erase(window);
+        } else {
+            ++window;
+        }
+        if (window == recreateWindows.end())
+            window = recreateWindows.begin();
+    }
+
+    mWaylandIntegration->reconfigureInputContext();
+}
+
 void QWaylandDisplay::flushRequests()
 {
     m_eventThread->readAndDispatchEvents();
@@ -425,6 +514,8 @@ void QWaylandDisplay::initEventThread()
             new EventThread(mDisplay, /* default queue */ nullptr, EventThread::EmitToDispatch));
     connect(m_eventThread.get(), &EventThread::needReadAndDispatch, this,
             &QWaylandDisplay::flushRequests, Qt::QueuedConnection);
+    connect(m_eventThread.get(), &EventThread::waylandError, this,
+            &QWaylandDisplay::checkWaylandError, Qt::QueuedConnection);
     m_eventThread->start();
 
     // wl_display_disconnect() free this.
@@ -434,10 +525,31 @@ void QWaylandDisplay::initEventThread()
     m_frameEventQueueThread->start();
 }
 
+void QWaylandDisplay::checkWaylandError()
+{
+    int ecode = wl_display_get_error(mDisplay);
+    if ((ecode == EPIPE || ecode == ECONNRESET)) {
+        qWarning("The Wayland connection broke. Did the Wayland compositor die?");
+        if (mWaylandTryReconnect) {
+            reconnect();
+            return;
+        }
+    } else {
+        qWarning("The Wayland connection experienced a fatal error: %s", strerror(ecode));
+    }
+    _exit(-1);
+}
+
 void QWaylandDisplay::blockingReadEvents()
 {
-    if (wl_display_dispatch(mDisplay) < 0)
-        checkWaylandError(mDisplay);
+    if (wl_display_dispatch(mDisplay) < 0) {
+        int ecode = wl_display_get_error(mDisplay);
+        if ((ecode == EPIPE || ecode == ECONNRESET))
+            qWarning("The Wayland connection broke during blocking read event. Did the Wayland compositor die?");
+        else
+            qWarning("The Wayland connection experienced a fatal error during blocking read event: %s", strerror(ecode));
+        _exit(-1);
+    }
 }
 
 void QWaylandDisplay::checkTextInputProtocol()
@@ -498,7 +610,7 @@ void QWaylandDisplay::registry_global(uint32_t id, const QString &interface, uin
 {
     struct ::wl_registry *registry = object();
 
-    static QByteArrayList interfaceBlacklist = qgetenv("QT_WAYLAND_DISABLED_INTERFACES").split(',');
+    static QStringList interfaceBlacklist = qEnvironmentVariable("QT_WAYLAND_DISABLED_INTERFACES").split(u',');
     if (interfaceBlacklist.contains(interface)) {
         return;
     }
@@ -514,62 +626,70 @@ void QWaylandDisplay::registry_global(uint32_t id, const QString &interface, uin
         mInputDevices.append(inputDevice);
 #if QT_CONFIG(wayland_datadevice)
     } else if (interface == QLatin1String(QWaylandDataDeviceManager::interface()->name)) {
-        mDndSelectionHandler.reset(new QWaylandDataDeviceManager(this, version, id));
+        mGlobals.dndSelectionHandler.reset(new QWaylandDataDeviceManager(this, version, id));
 #endif
     } else if (interface == QLatin1String(QtWayland::qt_surface_extension::interface()->name)) {
-        mWindowExtension.reset(new QtWayland::qt_surface_extension(registry, id, 1));
+        mGlobals.surfaceExtension.reset(new QtWayland::qt_surface_extension(registry, id, 1));
     } else if (interface == QLatin1String(QtWayland::wl_subcompositor::interface()->name)) {
-        mSubCompositor.reset(new QtWayland::wl_subcompositor(registry, id, 1));
+        mGlobals.subCompositor.reset(new QtWayland::wl_subcompositor(registry, id, 1));
     } else if (interface == QLatin1String(QWaylandTouchExtension::interface()->name)) {
-        mTouchExtension.reset(new QWaylandTouchExtension(this, id));
+        mGlobals.touchExtension.reset(new QWaylandTouchExtension(this, id));
     } else if (interface == QLatin1String(QWaylandQtKeyExtension::interface()->name)) {
-        mQtKeyExtension.reset(new QWaylandQtKeyExtension(this, id));
+        mGlobals.qtKeyExtension.reset(new QWaylandQtKeyExtension(this, id));
 #if QT_CONFIG(tabletevent)
     } else if (interface == QLatin1String(QWaylandTabletManagerV2::interface()->name)) {
-        mTabletManager.reset(new QWaylandTabletManagerV2(this, id, qMin(1, int(version))));
+        mGlobals.tabletManager.reset(new QWaylandTabletManagerV2(this, id, qMin(1, int(version))));
 #endif
     } else if (interface == QLatin1String(QWaylandPointerGestures::interface()->name)) {
-        mPointerGestures.reset(new QWaylandPointerGestures(this, id, 1));
+        mGlobals.pointerGestures.reset(new QWaylandPointerGestures(this, id, 1));
 #if QT_CONFIG(wayland_client_primary_selection)
     } else if (interface == QLatin1String(QWaylandPrimarySelectionDeviceManagerV1::interface()->name)) {
-        mPrimarySelectionManager.reset(new QWaylandPrimarySelectionDeviceManagerV1(this, id, 1));
+        mGlobals.primarySelectionManager.reset(
+                new QWaylandPrimarySelectionDeviceManagerV1(this, id, 1));
         for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
-            inputDevice->setPrimarySelectionDevice(mPrimarySelectionManager->createDevice(inputDevice));
+            inputDevice->setPrimarySelectionDevice(
+                    mGlobals.primarySelectionManager->createDevice(inputDevice));
 #endif
     } else if (interface == QLatin1String(QtWayland::qt_text_input_method_manager_v1::interface()->name)
             && (mTextInputManagerList.contains(interface) && mTextInputManagerList.indexOf(interface) < mTextInputManagerIndex)) {
         qCDebug(lcQpaWayland) << "text input: register qt_text_input_method_manager_v1";
         if (mTextInputManagerIndex < INT_MAX) {
-            mTextInputManagerv1.reset();
-            mTextInputManagerv2.reset();
+            mGlobals.textInputManagerv1.reset();
+            mGlobals.textInputManagerv2.reset();
 #if QT_WAYLAND_TEXT_INPUT_V4_WIP
-            mTextInputManagerv4.reset();
+            mGlobals.textInputManagerv4.reset();
 #endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
             for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                 inputDevice->setTextInput(nullptr);
         }
 
-        mTextInputMethodManager.reset(new QtWayland::qt_text_input_method_manager_v1(registry, id, 1));
+        mGlobals.textInputMethodManager.reset(
+                new QtWayland::qt_text_input_method_manager_v1(registry, id, 1));
         for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
-            inputDevice->setTextInputMethod(new QWaylandTextInputMethod(this, mTextInputMethodManager->get_text_input_method(inputDevice->wl_seat())));
+            inputDevice->setTextInputMethod(new QWaylandTextInputMethod(
+                    this,
+                    mGlobals.textInputMethodManager->get_text_input_method(
+                            inputDevice->wl_seat())));
         mWaylandIntegration->reconfigureInputContext();
         mTextInputManagerIndex = mTextInputManagerList.indexOf(interface);
     } else if (interface == QLatin1String(QtWayland::zwp_text_input_manager_v1::interface()->name)
                && (mTextInputManagerList.contains(interface) && mTextInputManagerList.indexOf(interface) < mTextInputManagerIndex)) {
         qCDebug(lcQpaWayland) << "text input: register zwp_text_input_v1";
         if (mTextInputManagerIndex < INT_MAX) {
-            mTextInputMethodManager.reset();
-            mTextInputManagerv2.reset();
+            mGlobals.textInputMethodManager.reset();
+            mGlobals.textInputManagerv2.reset();
 #if QT_WAYLAND_TEXT_INPUT_V4_WIP
-            mTextInputManagerv4.reset();
+            mGlobals.textInputManagerv4.reset();
 #endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
             for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                 inputDevice->setTextInputMethod(nullptr);
         }
 
-        mTextInputManagerv1.reset(new QtWayland::zwp_text_input_manager_v1(registry, id, 1));
+        mGlobals.textInputManagerv1.reset(
+                new QtWayland::zwp_text_input_manager_v1(registry, id, 1));
         for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices)) {
-            auto textInput = new QWaylandTextInputv1(this, mTextInputManagerv1->create_text_input());
+            auto textInput =
+                    new QWaylandTextInputv1(this, mGlobals.textInputManagerv1->create_text_input());
             textInput->setSeat(inputDevice->wl_seat());
             inputDevice->setTextInput(textInput);
         }
@@ -580,18 +700,20 @@ void QWaylandDisplay::registry_global(uint32_t id, const QString &interface, uin
             && (mTextInputManagerList.contains(interface) && mTextInputManagerList.indexOf(interface) < mTextInputManagerIndex)) {
         qCDebug(lcQpaWayland) << "text input: register zwp_text_input_v2";
         if (mTextInputManagerIndex < INT_MAX) {
-            mTextInputMethodManager.reset();
-            mTextInputManagerv1.reset();
+            mGlobals.textInputMethodManager.reset();
+            mGlobals.textInputManagerv1.reset();
 #if QT_WAYLAND_TEXT_INPUT_V4_WIP
-            mTextInputManagerv4.reset();
+            mGlobals.textInputManagerv4.reset();
 #endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
             for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                 inputDevice->setTextInputMethod(nullptr);
         }
 
-        mTextInputManagerv2.reset(new QtWayland::zwp_text_input_manager_v2(registry, id, 1));
+        mGlobals.textInputManagerv2.reset(
+                new QtWayland::zwp_text_input_manager_v2(registry, id, 1));
         for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
-            inputDevice->setTextInput(new QWaylandTextInputv2(this, mTextInputManagerv2->get_text_input(inputDevice->wl_seat())));
+            inputDevice->setTextInput(new QWaylandTextInputv2(
+                    this, mGlobals.textInputManagerv2->get_text_input(inputDevice->wl_seat())));
         mWaylandIntegration->reconfigureInputContext();
         mTextInputManagerIndex = mTextInputManagerList.indexOf(interface);
 #if QT_WAYLAND_TEXT_INPUT_V4_WIP
@@ -599,38 +721,48 @@ void QWaylandDisplay::registry_global(uint32_t id, const QString &interface, uin
             && (mTextInputManagerList.contains(interface) && mTextInputManagerList.indexOf(interface) < mTextInputManagerIndex)) {
         qCDebug(lcQpaWayland) << "text input: register zwp_text_input_v4";
         if (mTextInputManagerIndex < INT_MAX) {
-            mTextInputMethodManager.reset();
-            mTextInputManagerv2.reset();
+            mGlobals.textInputMethodManager.reset();
+            mGlobals.textInputManagerv2.reset();
             for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                 inputDevice->setTextInputMethod(nullptr);
         }
 
-        mTextInputManagerv4.reset(new QtWayland::zwp_text_input_manager_v4(registry, id, 1));
+        mGlobals.textInputManagerv4.reset(
+                new QtWayland::zwp_text_input_manager_v4(registry, id, 1));
         for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
-            inputDevice->setTextInput(new QWaylandTextInputv4(this, mTextInputManagerv4->get_text_input(inputDevice->wl_seat())));
+            inputDevice->setTextInput(new QWaylandTextInputv4(
+                    this, mGlobals.textInputManagerv4->get_text_input(inputDevice->wl_seat())));
         mWaylandIntegration->reconfigureInputContext();
         mTextInputManagerIndex = mTextInputManagerList.indexOf(interface);
 #endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
     } else if (interface == QLatin1String(QWaylandHardwareIntegration::interface()->name)) {
         bool disableHardwareIntegration = qEnvironmentVariableIntValue("QT_WAYLAND_DISABLE_HW_INTEGRATION");
         if (!disableHardwareIntegration) {
-            mHardwareIntegration.reset(new QWaylandHardwareIntegration(registry, id));
+            mGlobals.hardwareIntegration.reset(new QWaylandHardwareIntegration(registry, id));
             // make a roundtrip here since we need to receive the events sent by
             // qt_hardware_integration before creating windows
             forceRoundTrip();
         }
     } else if (interface == QLatin1String(QWaylandXdgOutputManagerV1::interface()->name)) {
-        mXdgOutputManager.reset(new QWaylandXdgOutputManagerV1(this, id, version));
+        mGlobals.xdgOutputManager.reset(new QWaylandXdgOutputManagerV1(this, id, version));
         for (auto *screen : std::as_const(mWaitingScreens))
             screen->initXdgOutput(xdgOutputManager());
     } else if (interface == QLatin1String(QtWayland::wp_fractional_scale_manager_v1::interface()->name)) {
-        mFractionalScaleManager.reset(new QtWayland::wp_fractional_scale_manager_v1(registry, id, 1));
+        mGlobals.fractionalScaleManager.reset(
+                new QtWayland::wp_fractional_scale_manager_v1(registry, id, 1));
     } else if (interface == QLatin1String("wp_viewporter")) {
-        mViewporter.reset(new QtWayland::wp_viewporter(registry, id, qMin(1u, version)));
+        mGlobals.viewporter.reset(new QtWayland::wp_viewporter(registry, id, qMin(1u, version)));
+    } else if (interface == QLatin1String(QtWayland::wp_cursor_shape_manager_v1::interface()->name)) {
+        mGlobals.cursorShapeManager.reset(
+                new QtWayland::wp_cursor_shape_manager_v1(registry, id, std::min(1u, version)));
+    } else if (
+            interface == QLatin1String(QtWayland::qt_toplevel_drag_manager_v1::interface()->name)) {
+        mGlobals.xdgToplevelDragManager.reset(
+                new QtWayland::qt_toplevel_drag_manager_v1(registry, id, 1));
     }
 
-    mGlobals.append(RegistryGlobal(id, interface, version, registry));
-    emit globalAdded(mGlobals.back());
+    mRegistryGlobals.append(RegistryGlobal(id, interface, version, registry));
+    emit globalAdded(mRegistryGlobals.back());
 
     const auto copy = mRegistryListeners; // be prepared for listeners unregistering on notification
     for (Listener l : copy)
@@ -639,8 +771,8 @@ void QWaylandDisplay::registry_global(uint32_t id, const QString &interface, uin
 
 void QWaylandDisplay::registry_global_remove(uint32_t id)
 {
-    for (int i = 0, ie = mGlobals.size(); i != ie; ++i) {
-        RegistryGlobal &global = mGlobals[i];
+    for (int i = 0, ie = mRegistryGlobals.size(); i != ie; ++i) {
+        RegistryGlobal &global = mRegistryGlobals[i];
         if (global.id == id) {
             if (global.interface == QLatin1String(QtWayland::wl_output::interface()->name)) {
                 for (auto *screen : mWaitingScreens) {
@@ -662,39 +794,39 @@ void QWaylandDisplay::registry_global_remove(uint32_t id)
                 }
             }
             if (global.interface == QLatin1String(QtWayland::zwp_text_input_manager_v1::interface()->name)) {
-                mTextInputManagerv1.reset();
+                mGlobals.textInputManagerv1.reset();
                 for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                     inputDevice->setTextInput(nullptr);
                 mWaylandIntegration->reconfigureInputContext();
             }
             if (global.interface == QLatin1String(QtWayland::zwp_text_input_manager_v2::interface()->name)) {
-                mTextInputManagerv2.reset();
+                mGlobals.textInputManagerv2.reset();
                 for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                     inputDevice->setTextInput(nullptr);
                 mWaylandIntegration->reconfigureInputContext();
             }
 #if QT_WAYLAND_TEXT_INPUT_V4_WIP
             if (global.interface == QLatin1String(QtWayland::zwp_text_input_manager_v4::interface()->name)) {
-                mTextInputManagerv4.reset();
+                mGlobals.textInputManagerv4.reset();
                 for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                     inputDevice->setTextInput(nullptr);
                 mWaylandIntegration->reconfigureInputContext();
             }
 #endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
             if (global.interface == QLatin1String(QtWayland::qt_text_input_method_manager_v1::interface()->name)) {
-                mTextInputMethodManager.reset();
+                mGlobals.textInputMethodManager.reset();
                 for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                     inputDevice->setTextInputMethod(nullptr);
                 mWaylandIntegration->reconfigureInputContext();
             }
 #if QT_CONFIG(wayland_client_primary_selection)
             if (global.interface == QLatin1String(QtWayland::zwp_primary_selection_device_manager_v1::interface()->name)) {
-                mPrimarySelectionManager.reset();
+                mGlobals.primarySelectionManager.reset();
                 for (QWaylandInputDevice *inputDevice : std::as_const(mInputDevices))
                     inputDevice->setPrimarySelectionDevice(nullptr);
             }
 #endif
-            emit globalRemoved(mGlobals.takeAt(i));
+            emit globalRemoved(mRegistryGlobals.takeAt(i));
             break;
         }
     }
@@ -702,7 +834,7 @@ void QWaylandDisplay::registry_global_remove(uint32_t id)
 
 bool QWaylandDisplay::hasRegistryGlobal(QStringView interfaceName) const
 {
-    for (const RegistryGlobal &global : mGlobals)
+    for (const RegistryGlobal &global : mRegistryGlobals)
         if (global.interface == interfaceName)
             return true;
 
@@ -713,8 +845,9 @@ void QWaylandDisplay::addRegistryListener(RegistryListener listener, void *data)
 {
     Listener l = { listener, data };
     mRegistryListeners.append(l);
-    for (int i = 0, ie = mGlobals.size(); i != ie; ++i)
-        (*l.listener)(l.data, mGlobals[i].registry, mGlobals[i].id, mGlobals[i].interface, mGlobals[i].version);
+    for (int i = 0, ie = mRegistryGlobals.size(); i != ie; ++i)
+        (*l.listener)(l.data, mRegistryGlobals[i].registry, mRegistryGlobals[i].id,
+                      mRegistryGlobals[i].interface, mRegistryGlobals[i].version);
 }
 
 void QWaylandDisplay::removeListener(RegistryListener listener, void *data)
@@ -790,6 +923,9 @@ void QWaylandDisplay::handleWindowDeactivated(QWaylandWindow *window)
 
     mActiveWindows.removeOne(window);
 
+    if (QCoreApplication::closingDown())
+        return;
+
     if (auto *decoration = window->decoration())
         decoration->update();
 }
@@ -859,7 +995,7 @@ void QWaylandDisplay::requestWaylandSync()
 
 QWaylandInputDevice *QWaylandDisplay::defaultInputDevice() const
 {
-    return mInputDevices.isEmpty() ? 0 : mInputDevices.first();
+    return mInputDevices.isEmpty() ? nullptr : mInputDevices.first();
 }
 
 bool QWaylandDisplay::isKeyboardAvailable() const
